#  Copyright (c) 2022 by Amplo.

import functools
import inspect
import logging
import re
import warnings
from collections.abc import Generator

__all__ = [
    "get_model",
    "hist_search",
    "clean_feature_name",
    "deprecated",
    "check_dtypes",
]


def get_model(model_str, **kwargs):
    # Import here to prevent ImportError (due to circular import)
    from amplo.automl.modelling import Modeller

    models = Modeller(**kwargs).return_models()
    model = [m for m in models if type(m).__name__ == model_str]

    try:
        return model[0]
    except IndexError:
        raise ValueError("Model not found.")


def hist_search(array, value):
    """
    Binary search that finds the index in order to fulfill
    ``array[index] <= value < array[index + 1]``

    Parameters
    ----------
    array : array of float
    value : float

    Returns
    -------
    int
        Bin index of the value
    """

    # Return -1 when no bin exists
    if value < array[0] or value >= array[-1]:
        logging.debug(
            f"No bin (index) found for value {value}. "
            f"Array(Min: {array[0]}, "
            "Max: {array[-1]})"
        )
        return -1

    # Initialize min and max bin index
    low = 0
    high = len(array) - 1

    # Bin search
    countdown = 30
    while countdown > 0:
        # Count down
        countdown -= 1

        # Set middle bin index
        middle = low + (high - low) // 2

        if low == middle == high - 1:  # stop criterion
            return middle

        if value < array[middle]:  # array[low] <= value < array[middle]
            high = middle
        elif value >= array[middle]:  # array[middle] <= value < array[high]
            low = middle

    warnings.warn("Operation took too long. Returning -1 (no match).", RuntimeWarning)
    return -1


def clean_feature_name(feature_name):
    """
    Clean feature names.

    Parameters
    ----------
    feature_name : string

    Returns
    -------
    cleaned_feature_name : string
    """
    # Remove non-numeric and non-alphabetic characters.
    # Assert single underscores and remove underscores in prefix and suffix.
    return re.sub("([^a-z0-9]|_)+", "_", str(feature_name).lower()).strip("_")


def deprecated(reason):
    # This decorator is a copy-pase from:
    # https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """
    string_types = (type(b""), type(""))

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                warnings.simplefilter("default", DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))


def check_dtypes(*dtype_tuples):
    """
    Checks all dtypes of given list.

    Parameters
    ----------
    *dtype_tuples : Any
        Tuples of (name, parameter, allowed types) to be checked.
        When checking only one parameter, the wrapping in a tuple can be omitted.

    Returns
    -------
    None

    Examples
    --------
    Check a single parameter:
    >>> check_dtypes(("var1", 123, int))  # tuple
    >>> check_dtypes("var1", 123, int)  # without tuple

    Check multiple:
    >>> check_dtypes(("var1", 123, int), ("var2", 1.0, (int, float)))  # tuples
    >>> check_dtypes(("var", var, str) for var in ["a", "b"])  # list or generator

    Raises
    ------
    ValueError
        If any given constraint is not fulfilled.
    """
    # Allow single dtype check without wrapping in a tuple
    if isinstance(dtype_tuples[0], str):
        if len(dtype_tuples) != 3:
            raise TypeError("Invalid arguments for `check_dtypes` function.")
        dtype_tuples = (dtype_tuples,)

    # Allow single list or generator object
    if isinstance(dtype_tuples[0], (list, Generator)):
        if len(dtype_tuples) != 1:
            raise TypeError("Invalid arguments for `check_dtypes` function.")
        dtype_tuples = dtype_tuples[0]

    # Check dtypes
    for name, value, typ in dtype_tuples:
        if not isinstance(value, typ):
            msg = f"Invalid dtype for argument `{name}`: " f"{type(value).__name__}"
            raise TypeError(msg)



#  Copyright (c) 2022 by Amplo.

import shutil
import time
from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification, make_regression

from amplo.automl import Modeller

__all__ = [
    "rmtree",
    "rmfile",
    "make_x_y",
    "make_data",
    "get_all_modeller_models",
    "find_first_warning_of_type",
    "RandomPredictor",
    "OverfitPredictor",
    "DelayedRandomPredictor",
]


def rmtree(folder="Auto_ML", must_exist=False):
    if Path(folder).exists():
        shutil.rmtree(folder)
    elif must_exist:
        raise FileNotFoundError(f"Directory {folder} does not exist")


def rmfile(file: str, must_exist=False):
    Path(file).unlink(missing_ok=not must_exist)


def make_x_y(mode: str):
    if mode == "classification":
        x, y = make_classification(n_features=5)
    elif mode == "multiclass":
        x, y = make_classification(n_features=5, n_classes=3, n_informative=3)
    elif mode == "regression":
        x, y = make_regression(n_features=5)
    else:
        raise ValueError("Invalid mode")
    x, y = pd.DataFrame(x), pd.Series(y)
    x.columns = [f"feature_{i}" for i in range(len(x.columns))]
    y.name = "target"
    return x, y


def make_data(mode: str, target="target"):
    data, y = make_x_y(mode)
    data[target] = y
    return data


def get_all_modeller_models(mode: str, **kwargs):
    models = {  # get each model type only once (!) with dictionary trick
        type(model).__name__: model
        for model in [
            *Modeller(mode=mode, samples=100, **kwargs).return_models(),
            *Modeller(mode=mode, samples=100_000, **kwargs).return_models(),
        ]
    }.values()
    return list(models)


def find_first_warning_of_type(typ, record):
    """
    Find first warning of type ``typ`` in warnings record.

    Parameters
    ----------
    typ : Any
        Warning to search for.
    record : pytest.WarningsRecorder
        Warnings record from ``with pytest.warns(...) as record`` context manager.

    Returns
    -------
    pytest.WarningsMessage
    """
    for item in record:
        if issubclass(item.category, typ):
            return item
    raise ValueError(f"No warning of type {typ} found in warnings record.")


# ----------------------------------------------------------------------
# Dummies


class _DummyPredictor:
    """
    Dummy predictor for testing.

    Parameters
    ----------
    mode : str
        Predicting mode ("classification" or "regression").
    """

    _dummy_classifier = None
    _dummy_regressor = None

    def __init__(self, mode):
        assert self._dummy_classifier is not None, "Dummy not set"
        assert self._dummy_regressor is not None, "Dummy not set"

        if mode == "classification":
            self.predictor = self._dummy_classifier()
        elif mode == "regression":
            self.predictor = self._dummy_regressor()
        else:
            raise ValueError("Invalid predictor mode.")

    def fit(self, x, y):
        return self.predictor.fit(x, y)

    def predict(self, x):
        return self.predictor.predict(x)

    def predict_proba(self, x):
        assert isinstance(self.predictor, self._dummy_classifier)
        return self.predictor.predict_proba(x)

    @property
    def classes_(self):
        if hasattr(self.predictor, "classes"):
            return self.predictor.classes


class _RandomClassifier:
    """
    Dummy classifier for testing.
    """

    def __init__(self):
        self.classes = None

    def fit(self, x, y):
        self.classes = np.unique(y)

    def predict(self, x):
        return np.random.choice(self.classes, len(x))

    def predict_proba(self, x):
        size = len(x), len(self.classes)
        proba = np.random.uniform(size=size)
        return proba * (1.0 / proba.sum(1)[:, np.newaxis])  # normalize


class _RandomRegressor:
    """
    Dummy regressor for testing.
    """

    def __init__(self):
        self.range = None

    def fit(self, x, y):
        self.range = np.min(y), np.max(y)

    def predict(self, x):
        return np.random.uniform(*self.range, len(x))


class RandomPredictor(_DummyPredictor):
    _dummy_classifier = _RandomClassifier
    _dummy_regressor = _RandomRegressor


class _OverfitClassifier:
    """
    Dummy classifier for testing. Returns the class if present in the data, else
    predicts 0
    """

    def __init__(self):
        self.classes = None
        self.x = None
        self.y = None

    def fit(self, x, y):
        self.x = x.to_numpy()
        self.y = y
        self.classes = y.unique()

    def predict(self, x):
        yt = []
        for i, row in x.iterrows():
            ind = np.where((row.values == self.x).all(axis=1))[0]
            if len(ind) == 0:
                yt.append(-1)
            else:
                yt.append(self.y.iloc[ind[0]])
        return yt

    def predict_proba(self, x):
        yt = []
        zeroes = [0 for _ in range(len(self.classes))]
        for i, row in x.iterrows():
            ind = np.where((row.values == self.x).all(axis=1))[0]
            if len(ind) == 0:
                yt.append(zeroes)
            else:
                yt.append(
                    [
                        0 if self.y.iloc[ind[0]] != i else 1
                        for i in range(len(self.classes))
                    ]
                )
        return yt


class _OverfitRegressor:
    """
    Dummy regressor for testing.
    """

    def __init__(self):
        self.classes = None
        self.x = None
        self.y = None

    def fit(self, x, y):
        self.x = x.to_numpy()
        self.y = y
        self.classes = y.unique()

    def predict(self, x):
        yt = []
        for i, row in x.iterrows():
            ind = np.where((row.values == self.x).all(axis=1))[0]
            if len(ind) == 0:
                yt.append(-1)
            else:
                yt.append(self.y.iloc[ind[0]])
        return yt


class OverfitPredictor(_DummyPredictor):
    _dummy_classifier = _OverfitClassifier
    _dummy_regressor = _OverfitRegressor


class DelayedRandomPredictor(RandomPredictor):
    def __init__(self, delay: float, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.delay = delay

    def predict(self, x):
        time.sleep(self.delay)
        return super().predict(x)

    def predict_proba(self, x):
        time.sleep(self.delay)
        return super().predict_proba(x)



#  Copyright (c) 2022 by Amplo.

import re
import warnings
from copy import deepcopy
from pathlib import Path

import pandas as pd

from amplo.api.platform import PlatformSynchronizer
from amplo.api.storage import AzureSynchronizer
from amplo.base.objects import LoggingMixin
from amplo.pipeline import Pipeline

__all__ = ["API"]


class API(LoggingMixin):
    def __init__(
        self,
        local_data_dir=None,
        download_data=True,
        upload_model=True,
        *,
        teams=None,
        machines=None,
        services=None,
        issues=None,
        verbose=1,
    ):
        """
        API for downloading data from blob storage, training models with AutoML and
        uploading them to Amplo`s platform.

        Parameters
        ----------
        local_data_dir : str or Path
            Destination directory for synchronizing data.
        download_data : bool
            Whether to synchronize from Azure blob storage.
        upload_model : bool
            Whether to upload trained model to platform.

        teams : list of str, optional
            Specify which teams will be considered for processing.
        machines : list of str, optional
            Specify which machines will be considered for processing.
        services : list of str, optional
            Specify which services will be considered for processing.
        issues : list of str, optional
            Specify which issues will be considered for processing.

        verbose : int
            Logging verbosity
        """
        super().__init__(verbose=verbose)

        if local_data_dir is None:
            warnings.warn(
                "No local data directory provided. Falling back to using current "
                "working directory."
            )
            local_data_dir = "."

        self.data_dir = Path(local_data_dir)
        self._download_data = bool(download_data)
        self._train_model = True
        self._upload_model = bool(upload_model)

        # Internal info
        self._trained_model_args = list()  # arguments for uploading model

        # Set up storage
        self.storage = AzureSynchronizer()
        self.platform = PlatformSynchronizer()

        # Set data arguments
        self.teams = self._set_azure_arg(teams)
        self.machines = self._set_azure_arg(machines)
        self.services = self._set_azure_arg(services)
        self.issues = self._set_azure_arg(issues)

    @staticmethod
    def _set_azure_arg(arg):
        """
        Put Azure argument into a list

        Parameters
        ----------
        arg : list or str or None

        Returns
        -------
        list of str
        """
        if isinstance(arg, list):
            return arg
        elif isinstance(arg, str):
            return [arg]
        elif arg is None:
            return ["*"]
        else:
            raise ValueError(
                f"List, string or None-type expected but got {type(arg)} instead."
            )

    def fit(self):
        self.sync_data()
        self.train_models()
        self.upload_models()

    def sync_data(self):
        """
        Synchronizes data from Azure blob storage

        Notes
        -----
        The algorithm iterates through all valid paths according to `self.teams`,
        `self.machines` and `self.services`. Note, however, that always all issues
        will be downloaded and cannot be specifically targeted
        """
        if not self._download_data:
            return

        self.logger.info("Downloading new data...")

        # Iterate blob data paths
        for issue_dir, team, machine, service, issue in self._iterate_data(
            from_local=False, level="issue"
        ):

            # Logging
            if self.verbose > 0:
                self.logger.info(
                    f"Downloading from: [Team] {team} - [Machine] {machine} - "
                    f"[Service] {service} - [Issue] {issue}"
                )

            # Define local saving directory (note the switch of `data`s position)
            local_issue_dir = str(
                self.data_dir / team / machine / service / "data" / issue
            )
            # Clean spaces in path
            local_issue_dir = re.sub(
                r"\s+", " ", local_issue_dir
            )  # double spaces to single spaces
            local_issue_dir = re.sub(
                r"^\s+", "", local_issue_dir
            )  # remove preceding spaces
            local_issue_dir = re.sub(
                r"\s+$", "", local_issue_dir
            )  # remove subsequent spaces

            # Synchronize all files of given issue
            self.storage.sync_files(
                issue_dir, local_issue_dir
            )  # TODO give `last_updated` argument

    def train_models(self, **kwargs):
        """
        Trains a model for every combination of
            a) team
            b) machine
            c) service
            d) issue

        Note that for the first 3 a selection can be specified in `self.__init__`.

        Parameters
        ----------
        **kwargs : optional
            Arguments to manipulate behavior of `Amplo.Pipeline`
        """

        if not self._train_model:
            return

        self.logger.info("Start model training...")

        for service_dir, team, machine, service in self._iterate_data(
            from_local=True, level="service"
        ):

            self.logger.info(
                f"Preparing data for: "
                f"[Team] {team} - [Machine] {machine} - [Service] {service}",
            )

            # Set up directories
            read_dir = service_dir / "data"
            model_dir = service_dir / "models"
            model_dir.mkdir(exist_ok=True)

            # --- Prepare common data (all models will act on the essentially same data)

            # Set pipeline arguments
            preparing_pipeline_kwargs = dict(
                target="labels",
                extract_features=False,
                balance=False,
                grid_search_time_budget=7200,
                grid_search_candidates=2,
            )
            preparing_pipeline_kwargs.update(kwargs)  # allow (optional) manipulation
            preparing_pipeline_kwargs["main_dir"] = f"{model_dir}/_Data_Preparation/"
            preparing_pipeline_kwargs[
                "name"
            ] = f"Preparing Pipeline ({team} - {machine} - {service})"

            # Check whether new data has been added
            checker = Pipeline(no_dirs=True, **preparing_pipeline_kwargs)
            if Path(checker.main_dir).exists():
                checker._load_version()  # noqa
            else:
                checker.version = 1
            # TODO: load settings directly, if present
            checker._read_data(read_dir)  # noqa
            if not checker.has_new_training_data():
                self.logger.info("Skipped training since no new data")
                continue

            # Set up pipeline and prepare data
            preparing_pipeline = Pipeline(**preparing_pipeline_kwargs)
            preparing_pipeline.data_preparation(read_dir)

            # Select issues to iterate
            all_issues = preparing_pipeline.y_orig.unique()
            iter_issues = all_issues if "*" in self.issues else self.issues

            # Train one model for each issue
            for issue in iter_issues:
                self.logger.info(f"Training model for {issue}")

                # Set up pipeline using a copy from the preparing pipeline
                pipeline = deepcopy(preparing_pipeline)
                pipeline.main_dir = f"{model_dir}/{issue}/"
                pipeline.name = f"{team} - {machine} - {service} - {issue}"

                # Create dirs
                if not pipeline.no_dirs:
                    pipeline._create_dirs()  # noqa
                    pipeline._load_version()  # noqa

                # Inject model specific labels
                new_y = pd.Series(pipeline.y_orig == issue, dtype="int32")
                pipeline.set_y(new_y)
                pipeline.n_classes = 2

                # Train models
                pipeline.model_training()
                pipeline.conclude_fitting()

                # Append to trained model arguments
                self._trained_model_args.append(
                    [pipeline.main_dir, team, machine, service, issue, pipeline.version]
                )

    def upload_models(self, model_args=None):
        """
        Upload trained models or, when `model_args` provided, upload given models.

        Parameters
        ----------
        model_args : typing.List[list or tuple], optional
            Model arguments. Each tuple contains following info in order:
                - issue directory [str or Path]
                - team [str]
                - machine [str]
                - service [str]
                - issue [str]
                - version [int]
        """

        if not self._upload_model and not model_args:
            # Skip if setting says not to upload trained models
            #  and no explicit model_args have been passed.
            return

        self.logger.info("Uploading trained models...")

        if model_args is not None and len(self._trained_model_args) > 0:
            warnings.warn(
                "An explicit list of model_args was provided despite > 1 model was "
                "trained.",
                UserWarning,
            )

        for model_args in model_args or self._trained_model_args:
            # Upload models to Amplo`s platform
            self.platform.upload_model(*model_args)

    # --- Utilities ---

    def _iterate_data(self, *, from_local=True, level="issue"):
        """
        Iterate over directories

        Parameters
        ----------
        from_local : bool
            Whether `parent_dir` is a local or a blob directory path
        level : str
            Select level of iterating data.
            One of ('team', 'machine', 'service', 'issue')

        Yields
        ------
        dir : Path
            Directory of data
        *info : str
            Names of team, machine, service and issue.
            Depends on the `level` parameter!
        """

        # Assertions
        assert level in ("team", "machine", "service", "issue"), "Unknown argument"

        # Init
        parent_dir = self.data_dir if from_local else None

        # Iterate teams
        for team_dir in self._iterate_directory(
            parent_dir, self.teams, from_local=from_local
        ):
            info = (team_dir.name,)

            if level == "team":
                yield team_dir, *info
                continue

            # Iterate machines
            for machine_dir in self._iterate_directory(
                team_dir, self.machines, from_local=from_local
            ):
                info = (team_dir.name, machine_dir.name)

                if level == "machine":
                    yield machine_dir, *info
                    continue

                # Adjust search path
                machine_dir_ = (
                    Path(f"{machine_dir}/data/") if not from_local else machine_dir
                )

                # Iterate services
                for service_dir in self._iterate_directory(
                    machine_dir_, self.services, from_local=from_local
                ):
                    info = (team_dir.name, machine_dir.name, service_dir.name)

                    if level == "service":
                        yield service_dir, *info
                        continue

                    # Adjust search path
                    service_dir_ = (
                        Path(f"{service_dir}/data/") if from_local else service_dir
                    )

                    # Iterate issues
                    for issue_dir in self._iterate_directory(
                        service_dir_, self.issues, from_local=from_local
                    ):
                        info = (
                            team_dir.name,
                            machine_dir.name,
                            service_dir.name,
                            issue_dir.name,
                        )

                        assert level == "issue"
                        yield issue_dir, *info

    def _iterate_directory(self, parent_dir=None, selection=None, from_local=True):
        """
        Iterate over subdirectories of parent directory.

        Parameters
        ----------
        parent_dir : str or Path, optional
            Parent directory.
        selection : str or list of str, optional
            Select subdirectories.
        from_local : bool
            Whether `parent_dir` is a local or a blob directory path.

        Notes
        -----
        The parameter `selection` recognizes RegEx flags. For example,
        ``selection="folder.*"`` would yield each subdirectory that begins with
        ``folder`` (``folder1``, ``folder_two``, ...).

        Yields
        ------
        Path
            Next subdirectory.
        """

        # Input checks
        if not isinstance(parent_dir, (str, Path)):
            err = ValueError(f"Invalid argument type: {type(parent_dir)}")
            if from_local:
                raise err
            elif parent_dir is not None:
                raise err
        if from_local and not Path(parent_dir).exists():
            raise FileNotFoundError(f"Directory does not exist: {parent_dir}")
        if selection is not None:
            if not hasattr(selection, "__iter__"):
                raise ValueError(f"Selection is not iterable: {selection}")
            elif not all(isinstance(sel, str) for sel in selection):
                raise ValueError("Selection must be an iterable of strings.")

        # Make selection regex-able
        if isinstance(selection, str):
            selection = [selection]
        if selection is None or any(sel in ("*", ".*") for sel in selection):
            selection = [".*"]
        else:
            selection = [sel + "$" for sel in selection]

        # Define sorted subdirectory iterator
        if from_local:
            subdirectories = sorted(
                child.resolve()
                for child in Path(parent_dir).iterdir()
                if child.is_dir()
            )
        else:
            subdirectories = sorted(
                Path(child) for child in self.storage.get_dir_paths(parent_dir)
            )

        # Yield matching subdirectories
        for subdir in subdirectories:
            # Note that matches for `re.match` must match at the beginning of a
            # given string. Otherwise, `re.find` would have been used.
            if any(re.match(match, subdir.name) for match in selection):
                yield subdir



#  Copyright (c) 2022 by Amplo.

import pandas as pd
import pytest
from sklearn.datasets import make_classification, make_regression
from sklearn.model_selection import KFold, StratifiedKFold

from amplo import Pipeline
from amplo.grid_search import ExhaustiveGridSearch, OptunaGridSearch
from amplo.utils.io import parse_json
from tests import get_all_modeller_models


@pytest.fixture(scope="class", params=["regression", "classification"])
def make_data(request):
    mode = request.param

    if mode == "regression":
        request.cls.k_fold = KFold
        request.cls.objective = "r2"
        x, y = make_regression(n_features=5)
    elif mode == "classification":
        request.cls.k_fold = StratifiedKFold
        request.cls.objective = "r2"
        x, y = make_classification(n_features=5)
    else:
        raise ValueError("Mode is invalid")

    request.cls.mode = mode
    request.cls.x = pd.DataFrame(x)
    request.cls.y = pd.Series(y)
    yield


@pytest.mark.usefixtures("make_data")
class TestGridSearch:
    @pytest.mark.parametrize(
        "test_case", ["grid_search_type", "grid_search_iterations"]
    )
    def test_no_grid_search(self, test_case):
        # Set keyword arguments
        kwargs = dict(extract_features=False, sequence=False, plot_eda=False)
        if test_case == "grid_search_type":
            kwargs.update({test_case: None})
        elif test_case == "grid_search_iterations":
            kwargs.update({test_case: 0})
        else:
            raise ValueError("Invalid test case")
        # Fit pipeline
        pipeline = Pipeline(**kwargs)
        pipeline.fit(self.x, self.y)
        # Check results
        assert any(
            pipeline.results.loc[:, "type"] == "Hyper Parameter"
        ), "No hyper parameter results were found"

    @pytest.mark.parametrize("grid_search", [ExhaustiveGridSearch, OptunaGridSearch])
    def test_each_model(self, grid_search):
        """
        Given the models which depend on 1) mode and 2) num_samples,
        each model is trained on the given grid search class.

        Thanks to the fact that all grid search models have the same interface,
        this is possible :-)
        """
        models = get_all_modeller_models(mode=self.mode, objective=self.objective)

        for model in models:
            # Grid search
            search = grid_search(
                model,
                cv=self.k_fold(n_splits=3),
                verbose=0,
                timeout=10,
                n_trials=2,
                scoring=self.objective,
            )
            results = search.fit(self.x, self.y)

            # Tests
            model_name = type(model).__name__
            assert isinstance(results, pd.DataFrame), (
                f"Expected result to be pandas.DataFrame"
                f"but found {type(results)} in {model_name} instead"
            )
            results_keys = [
                "worst_case",
                "mean_objective",
                "std_objective",
                "params",
                "mean_time",
                "std_time",
            ]
            assert all(
                key in results.keys() for key in results_keys
            ), f"Keys are missing in {model_name}"
            assert len(results) > 0, f"Results are empty in {model_name}"
            model.set_params(**parse_json(results.iloc[0]["params"]))



#  Copyright (c) 2022 by Amplo.

import json
import os
import re
import warnings
from pathlib import Path

import requests

from amplo.utils import check_dtypes

__all__ = ["PlatformSynchronizer"]


class PlatformSynchronizer:
    def __init__(
        self,
        connection_string_name="AMPLO_PLATFORM_STRING",
        api_url="https://platform.amplo.ch/api",
        verbose=0,
    ):
        """
        Connector to Amplo`s platform for uploading trained models

        Parameters
        ----------
        connection_string_name : str
        api_url : str
        verbose : int
        """
        self.api_url = re.sub(r"/$", "", api_url)
        self.api_key = os.getenv(connection_string_name)
        self.verbose = int(verbose)

    def upload_model(self, issue_dir, team, machine, service, issue, version, model_id):
        """
        Upload trained model

        Parameters
        ----------
        issue_dir : str or Path
            Model issue directory
        team : str
            Team name
        machine : str
            Machine name
        service : str
            Service name
        issue : str
            Issue name
        version : int
            Version of production model to upload
        model_id : int
            Model identifier for platform.

        Notes
        -----
        Data structures
            - Locally: ``Auto_ML / Production / vX / ...``
            - Cloud: ``Team / Machine / models / Diagnostics / vX / ...``
        """

        # Check input args
        if not os.path.isdir(issue_dir):
            raise FileNotFoundError("Model directory does not exist.")
        check_dtypes(
            ("team", team, str),
            ("machine", machine, str),
            ("service", service, str),
            ("issue", issue, str),
            ("version", version, int),
        )

        # Check file structure
        file_dir = Path(issue_dir) / "Production" / f"v{version}"
        if not os.path.isdir(file_dir):
            raise FileNotFoundError(
                f"Production files for version {version} don't exist."
            )
        files = [file_dir / file for file in ("Settings.json", "Model.joblib")]
        for file in files:
            if not file.exists():
                raise FileNotFoundError(f"File {file} doesn't exist in {file_dir}.")

        # Set info
        headers = {"X-Api-Key": self.api_key}
        data = {
            "team": team,  # team name
            "machine": machine,  # machine name
            "category": service,  # category
            "model": issue,
            "name": issue,  # model name (for PUT or POST, respectively)
            "version": f"v{version}",  # latest version
        }

        # Assert that model exists
        def model_exists(model_name: str):
            # GET request
            get_response = requests.get(
                self.api_url + "/models", params=data, headers=headers
            )
            if get_response.status_code != 200:
                raise requests.HTTPError(f"{get_response} {get_response.text}")

            # Get all existing models
            try:
                # Convert string to list of dict
                existing_models = json.loads(get_response.text)
                if isinstance(existing_models, str):
                    ValueError("Could not transform response.")
            except json.decoder.JSONDecodeError as err:
                warnings.warn(f"Could not decode GET response - {err}")
                # Set default value of ``existing_models`` to list of (empty) dict
                existing_models = list(dict())

            # Check existing models
            if not isinstance(existing_models, list):
                raise ValueError(f"Unexpected data type: {type(existing_models)}")
            for model in existing_models:
                if not isinstance(model, dict):
                    raise ValueError(f"Unexpected data type: {type(model)}")

            # Check if model exists
            return any(
                model.get("name", None) == model_name for model in existing_models
            )

        # POST when model yet doesn't exist
        if not model_exists(data["name"]):
            warnings.warn(
                f"Model {data['name']} did not yet exist. POST-ing a new model to the "
                f"platform..."
            )
            post_response = requests.post(
                self.api_url + "/models", data=data, headers=headers
            )
            # Assert that now model exists
            if not model_exists(data["name"]):
                raise requests.HTTPError(f"{post_response} {post_response.text}")

        # Add files and model id to data
        data.update({"files": [open(file, "r") for file in files], "id": model_id})
        # PUT a new version
        put_response = requests.put(
            self.api_url + "/trainings", data=data, headers=headers
        )
        if put_response.status_code != 200:
            raise requests.HTTPError(f"{put_response} {put_response.text}")

    def upload_latest_model(self, issue_dir, *args, **kwargs):
        """
        Same as `upload_model` but selects the latest version of production model

        Parameters
        ----------
        issue_dir : str or Path
            Model issue directory
        args
            Arguments to be passed to `upload_model`
        kwargs
            Keyword arguments to be passed to `upload model`
        """
        # Check input args
        assert os.path.isdir(issue_dir), "Model directory does not exist"

        # Check file structure
        all_version_dirs = [
            dir_
            for dir_ in (Path(issue_dir) / "Production").glob("v*")
            if dir_.is_dir()
        ]
        assert len(all_version_dirs) > 0, "No production data found"

        # Take the latest version and remove `v` in e.g. `v1`
        latest_version = int(sorted(all_version_dirs)[-1].name[1:])

        # Set version in kwargs and upload model
        kwargs.update({"version": latest_version})
        return self.upload_model(issue_dir, *args, **kwargs)



#  Copyright (c) 2022 by Amplo.

import json
import re

import numpy as np
import pandas as pd
import pytest

from amplo import Pipeline
from amplo.observation._base import ProductionWarning
from amplo.observation.data import DataObserver
from tests import find_first_warning_of_type


class TestDataObserver:
    @pytest.mark.parametrize("index_type", ["single", "multi"])
    def test_monotonic_columns(self, index_type):
        size = 100
        monotonic_incr = 4.2 * np.arange(-10, size - 10)[:, None]  # start=-10, step=4.2
        monotonic_decr = 6.3 * np.arange(-3, size - 3)[::-1, None]
        constants = np.zeros(size)[:, None]
        random = np.random.normal(size=size)[:, None]

        # Prepare data
        x1 = np.concatenate([monotonic_incr, monotonic_decr, constants, random], axis=1)
        x2 = np.concatenate(4 * [random], axis=1)
        y1 = random.reshape(-1)  # does not matter
        if index_type == "single":
            x = pd.DataFrame(x1)
            y = pd.Series(y1)
        elif index_type == "multi":
            x = pd.DataFrame(np.concatenate([x1, x2], axis=0))
            x.index = pd.MultiIndex.from_product([[0, 1], range(size)])
            y = pd.Series(np.concatenate([y1, y1], axis=0))
            y.index = x.index
        else:
            raise ValueError("Invalid index_type.")

        # Add nan to first and random
        x.iloc[1, 0] = np.nan

        # Observe
        pipeline = Pipeline(n_grid_searches=0)
        pipeline._read_data(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_monotonic_columns()
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        monotonic_cols = json.loads(re.search(r"\[.*]", msg).group(0))
        assert set(monotonic_cols) == {0, 1}, "Wrong monotonic columns identified."

    def test_minority_sensitivity(self):
        # Setup
        x = np.hstack(
            (
                np.random.normal(size=(100, 1)),
                np.concatenate((np.zeros((2, 1)), np.random.normal(100, 1, (98, 1)))),
            )
        )
        y = np.concatenate((np.zeros(5), np.ones(95)))

        # Add nan
        x[1, 0] = np.nan

        # Observe
        pipeline = Pipeline(n_grid_searches=0)
        pipeline._read_data(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_minority_sensitivity()
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        sensitive_cols = json.loads(re.search(r"\[.*]", msg).group(0))
        assert sensitive_cols == [1], "Wrong minority sensitive columns identified."

    def test_categorical_mismatch(self):
        # Setup
        x = np.hstack(
            (
                np.array(["New York"] * 50 + ["new-york"] * 50).reshape((-1, 1)),
                np.array(["Something"] * 50 + ["Somethign"] * 50).reshape((-1, 1)),
                np.random.normal(100, 5, (100, 1)),
            )
        )
        y = np.concatenate((np.zeros(5), np.ones(95)))

        # Add nan
        x[0, 0] = np.nan

        # Observe
        pipeline = Pipeline(n_grid_searches=0, cv_splits=3)
        pipeline.data_preparation(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_categorical_mismatch()
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        sensitive_cols = json.loads(re.search(r"\[.*]", msg).group(0))
        assert sensitive_cols == [
            {"feature_1": ["something", "somethign"]}
        ], "Wrong categorical mismatch columns identified."

    def test_extreme_values(self):
        # Setup
        x = np.vstack((np.random.normal(size=100), np.linspace(1000, 10000, 100))).T
        y = np.concatenate((np.zeros(5), np.ones(95)))

        # Add nan
        x[0, 0] = np.nan

        # Observe
        pipeline = Pipeline(n_grid_searches=0)
        pipeline._read_data(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_extreme_values()
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        extreme_cols = json.loads(re.search(r"\[.*]", msg).group(0))
        assert extreme_cols == [1], "Wrong minority sensitive columns identified."

    def test_label_issues(self):
        # Setup
        x = np.hstack(
            (np.random.normal(-2, 0.1, size=50), np.random.normal(2, 0.1, size=50))
        )
        y = np.hstack((np.ones(50), np.zeros(49), 1))

        # Observe
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(x, y)
        obs = DataObserver(pipeline=pipeline)
        with pytest.warns(ProductionWarning) as record:
            obs.check_label_issues()
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        label_issues = json.loads(re.search(r"\[.*]", msg).group(0))
        assert label_issues == [99], "Wrong label issues identified."



#  Copyright (c) 2022 by Amplo.

from copy import deepcopy

from amplo.regression._base import BaseRegressor


class PartialBoostingRegressor(BaseRegressor):
    """
    Amplo wrapper for regressor boosting models.

    The number of estimators being used in the prediction are limited.

    Parameters
    ----------
    model
        Boosting model to wrap.
    step : int
        Number of iterations/estimators to limit the model on predictions.
    verbose : {0, 1, 2}
        Verbose logging.
    """

    _SUPPORTED_MODELS = [
        "AdaBoostRegressor",
        "GradientBoostingRegressor",
        "LGBMRegressor",
        "XGBRegressor",
        "CatBoostRegressor",
    ]

    def __init__(self, model, step, verbose=0):
        model = deepcopy(model)

        model_name = type(model).__name__
        if model_name not in self._SUPPORTED_MODELS:
            raise ValueError(f"Unsupported model {model_name}")
        if model_name in ("AdaBoostRegressor", "GradientBoostingRegressor"):
            model.estimators_ = model.estimators_[:step]

        super().__init__(model=model, verbose=verbose)
        self.step = int(step)

    def _get_prediction_kwargs(self):
        model_name = type(self.model).__name__
        if model_name in ("AdaBoostRegressor", "GradientBoostingRegressor"):
            return {}
        elif model_name == "LGBMRegressor":
            return {"num_iterations": self.step}
        elif model_name == "XGBRegressor":
            return {"iteration_range": (0, self.step)}
        elif model_name == "CatBoostRegressor":
            return {"ntree_end": self.step}
        else:
            raise AttributeError(f"Unsupported model {model_name}")

    def _predict(self, x, y=None, **kwargs):
        return self.model.predict(x, **kwargs, **self._get_prediction_kwargs())

    @classmethod
    def n_estimators(cls, model):
        model_name = type(model).__name__
        if model_name in ("AdaBoostRegressor", "GradientBoostingRegressor"):
            return len(model.estimators_)
        elif model_name in ("LGBMRegressor", "XGBRegressor"):
            return model.model.n_estimators
        elif model_name == "CatBoostRegressor":
            return model.model.tree_count_
        else:
            raise ValueError(f"Unsupported model {model_name}")



#  Copyright (c) 2022 by Amplo.

import multiprocessing as mp
import warnings
from datetime import datetime

import numpy as np
import pandas as pd
from scipy.stats import loguniform, randint, uniform
from sklearn.experimental import enable_halving_search_cv  # noqa
from sklearn.model_selection import HalvingRandomSearchCV  # noqa

from amplo.grid_search._base import BaseGridSearch
from amplo.utils import deprecated


@deprecated("Please use Optuna, not experimental and more efficient.")
class HalvingGridSearch(BaseGridSearch):
    def __init__(self, model, *args, **kwargs):
        if "timeout" in kwargs:
            warnings.warn("Parameter `timeout` has no effect")
            kwargs.pop("timeout")

        super().__init__(model, *args, **kwargs)

        # Halving settings
        self.resource = "n_samples"
        self.max_resource = "auto"
        self.min_resource = 200
        self.set_resources()

    def set_resources(self):
        if "CatBoost" in type(self.model).__name__:
            self.resource = "n_estimators"
            self.max_resource = 3000
            self.min_resource = 250
        if (
            self.model.__module__ == "sklearn.ensemble._bagging"
            or self.model.__module__ == "xgboost.sklearn"
            or self.model.__module__ == "lightgbm.sklearn"
            or self.model.__module__ == "sklearn.ensemble._forest"
        ):
            self.resource = "n_estimators"
            self.max_resource = 1500
            self.min_resource = 50

    def _get_hyper_params(self):
        param_values = self._hyper_parameter_values
        param_values.pop(
            "CONDITIONALS", {}
        )  # drop conditionals as they are not supported
        param_values.pop(
            "n_jobs", None
        )  # drop n_jobs as HalvingGridSearch also uses this parameter
        param_values.pop(
            "n_estimators", None
        )  # otherwise cannot use parameter n_estimators as the resource

        # Drop/adjust some more parameter names which HalvinGridSearch cannot handle
        model_name = type(self.model).__name__
        if model_name in ("BaggingClassifier", "BaggingRegressor"):
            # Apparently the regressor cannot handle any features for halving
            return {}
        elif model_name in ("LGBMRegressor", "LGBMClassifier"):
            # Someway it cannot handle uniform distributions for this two cases
            param_values["bagging_fraction"] = (
                "categorical",
                [0.5, 0.7, 0.9, 1.0],
                None,
            )
            param_values["feature_fraction"] = (
                "categorical",
                [0.5, 0.7, 0.9, 1.0],
                None,
            )

        # Extract parameter distributions
        params = {}
        for p_name, value in param_values.items():

            # Read out
            p_type = value[0]  # parameter type (str)
            p_args = value[1]  # parameter arguments (list, tuple)

            # Sanity checks
            assert (
                len(p_args) == 2 or p_type == "categorical"
            ), "Only categorical parameter can have more/less than two suggest args"

            # Suggest parameter given the arguments
            if p_type == "categorical":
                params[p_name] = p_args
            elif p_type == "int":
                params[p_name] = randint(*p_args)
            elif p_type == "logint":
                raise NotImplementedError("logint is not yet implemented")
            elif p_type == "uniform":
                params[p_name] = uniform(*p_args)
            elif p_type == "loguniform":
                params[p_name] = loguniform(*p_args)
            else:
                raise NotImplementedError("Invalid parameter specification")

        return params

    def fit(self, x, y):
        # Update minimum resource for samples (based on dataset)
        if self.resource == "n_samples":
            self.min_resource = int(0.2 * len(x)) if len(x) > 5000 else len(x)

        # Parameters
        if len(np.unique(y)) == 2:
            self.binary = True
        self.samples = len(y)
        if self.params is None:
            self.params = self._get_hyper_params()

        # Set up and run grid search
        halving_random_search = HalvingRandomSearchCV(
            self.model,
            self.params,
            n_candidates=self.n_trials,
            resource=self.resource,
            max_resources=self.max_resource,
            min_resources=self.min_resource,
            cv=self.cv,
            scoring=self.scoring,
            factor=3,
            n_jobs=mp.cpu_count() - 1,
            verbose=self.verbose,
        )
        halving_random_search.fit(x, y)

        # Parse results
        scikit_results = pd.DataFrame(halving_random_search.cv_results_)
        results = pd.DataFrame(
            {
                "date": datetime.today().strftime("%d %b %y"),
                "model": type(self.model).__name__,
                "params": scikit_results["params"],
                "mean_objective": scikit_results["mean_test_score"],
                "std_objective": scikit_results["std_test_score"],
                "worst_case": (
                    scikit_results["mean_test_score"] - scikit_results["std_test_score"]
                ),
                "mean_time": scikit_results["mean_fit_time"],
                "std_time": scikit_results["std_fit_time"],
            }
        )

        # Update resource in results
        if self.resource != "n_samples":
            for i in range(len(results)):
                results.loc[results.index[i], "params"][
                    self.resource
                ] = self.max_resource

        return results



#  Copyright (c) 2022 by Amplo.

"""
Enables connection to Databricks via API calls.
"""


from __future__ import annotations

import json
import os
from copy import deepcopy

import requests

from amplo.utils import check_dtypes

__all__ = ["DatabricksJobsAPI", "train_on_cloud"]


class DatabricksJobsAPI:
    """
    Helper class for working with Databricks' Job API.

    References
    ----------
    [API reference](https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/jobs)

    Parameters
    ----------
    host : str
        Databricks host (see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/authentication).
    access_token : str
        Access token (see Databricks > User Settings > Access tokens).
    """

    def __init__(self, host: str, access_token: str):
        check_dtypes(("host", host, str), ("access_token", access_token, str))
        self.host = host
        self.access_token = access_token

    @classmethod
    def from_os_env(
        cls, host: str = None, access_token: str = None
    ) -> DatabricksJobsAPI:
        """
        Instantiate the class using os environment strings.

        Parameters
        ----------
        host : str, default: "DATABRICKS_HOST"
            Key in the os environment for the Databricks host.
        access_token : str, default: "DATABRICKS_TOKEN"
            Key in the os environment for the Databricks access token.

        Returns
        -------
        DatabricksJobsAPI
        """
        host = host or "DATABRICKS_HOST"
        access_token = access_token or "DATABRICKS_TOKEN"
        check_dtypes(("host", host, str), ("access_token", access_token, str))
        return cls(os.getenv(host), os.getenv(access_token))

    def __repr__(self):
        """
        Readable string representation of the class.
        """

        return f"{self.__class__.__name__}: <{self.host}>"

    def request(self, method: str, action: str, body: dict = None) -> dict:
        """
        Send a request to Databricks.

        Notes
        -----
        Every Databricks Jobs API call can be requested from here. In case you need more
        functionality than already implemented, have a look at the [OpenAPI
        specification](https://docs.microsoft.com/en-us/azure/databricks/_static/api-refs/jobs-2.1-azure.yaml).

        Parameters
        ----------
        method : str
            Request method (``GET``, ``PUT``, ``POST``, ...).
        action : str
            Path to Databricks Job (see Databricks Jobs API specification).
        body : dict of {str: int or str}, optional
            Request body schema (see Databricks Jobs API specification).

        Returns
        -------
        dict or {str: int or str}
            The request's response.

        Raises
        ------
        requests.HTTPError
            When the request's response has another status code than 200.
        """

        url = f"{self.host}/api/{action.lstrip('/')}"
        headers = {"Authorization": f"Bearer {self.access_token}"}
        body = body or {}

        response = requests.request(method, url, headers=headers, json=body)
        if not response.status_code == 200:
            raise requests.HTTPError(f"{response} {response.text}")

        return response.json()

    # --------------------------------------------------------------------------
    # Jobs API requests

    def list_jobs(
        self, limit: int = None, offset: int = None, expand_tasks: bool = None
    ) -> dict:
        body = {"limit": limit, "offset": offset, "expand_tasks": expand_tasks}
        return self.request("get", "2.1/jobs/list", body)

    def list_runs(
        self,
        limit: int = None,
        offset: int = None,
        expand_tasks: bool = None,
        *,
        active_only: bool = None,
        completed_only: bool = None,
        job_id: int = None,
        start_time_from: int = None,
        start_time_to: int = None,
    ) -> dict:
        body = {
            "limit": limit,
            "offset": offset,
            "expand_tasks": expand_tasks,
            "active_only": active_only,
            "completed_only": completed_only,
            "job_id": job_id,
            "start_time_from": start_time_from,
            "start_time_to": start_time_to,
        }
        return self.request("get", "2.1/jobs/runs/list", body)

    def get_job(self, job_id: int) -> dict:
        return self.request("get", "2.1/jobs/get", {"job_id": job_id})

    def get_run(self, run_id: int, include_history: bool = None) -> dict:
        body = {"run_id": run_id, include_history: include_history}
        return self.request("get", "2.1/jobs/runs/get", body)

    def run_job(
        self,
        job_id: int,
        idempotency_token: str = None,
        notebook_params: dict[str, int | str] = None,
        # Note: more parameters are available
    ) -> dict:
        body = {
            "job_id": job_id,
            "idempotency_token": idempotency_token,
            "notebook_params": notebook_params,
        }
        return self.request("post", "2.1/jobs/run-now", body)

    def cancel_run(self, run_id: int) -> dict:
        return self.request("post", "2.1/jobs/runs/cancel", {"run_id": run_id})


def train_on_cloud(
    job_id: int,
    model_version: int = 1,
    idempotency_token: str = None,
    notebook_params: dict[str, dict | bool | int | str] = None,
    *,
    host_key: str = None,
    access_token_key: str = None,
) -> dict[str, int]:
    """
    Starts a job run via the DatabricksJobsAPI.

    Notes
    -----
    Make sure to have set the following environment variables:
        - ``DATABRICKS_INSTANCE``
          (see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/authentication).
        - ``DATABRICKS_ACCESS_TOKEN`` (see Databricks > User Settings > Access tokens).

    Note two important differences to ``DatabricksJobsAPI.run_job``.
    The "pipe_kwargs" key of ``notebook_params``:
        - will be JSON dumped to a string for you.
        - gets default values imputed if not given.

    Parameters
    ----------
    job_id : int
        Job ID in Databricks.
    model_version : int, default: 1
        Version of the model. Will overwrite the "version" key in the "pipe_kwargs" key
        of ``notebook_params``.
    idempotency_token : str, optional
        Idempotency token to ensure uniqueness of job run.
    notebook_params : dict of {str: bool or int or str}
        Parameters for job run. See widgets in notebook of corresponding job.
    *
    host_key : str, default: "DATABRICKS_HOST"
        Key in the os environment for the Databricks host.
    access_token_key : str, default: "DATABRICKS_TOKEN"
        Key in the os environment for the Databricks access token.

    Returns
    -------
    dict of {str: int}
        If response is success (200), ``run_id`` (globally unique key of newly triggered
        run) is one of the present keys.

    Raises
    ------
    ValueError
        When one or more widget / keywords in ``notebook_params`` is missing.
    """

    check_dtypes(
        ("job_id", job_id, int),
        ("model_version", model_version, int),
        ("idempotency_token", idempotency_token, (type(None), str)),
        ("notebook_params", notebook_params, dict),
    )
    notebook_params = deepcopy(notebook_params)

    # Define default keyword arguments and overwrite them when given by user input.
    pipe_kwargs = {
        "interval_analyse": False,
        "standardize": False,
        "missing_values": "zero",
        "balance": False,
        "stacking": False,
        "grid_search_time_budget": 7200,
        "n_grid_searches": 1,
        "verbose": 1,
    }
    pipe_kwargs.update(notebook_params.get("pipe_kwargs", {}))
    pipe_kwargs["version"] = model_version
    notebook_params["pipe_kwargs"] = json.dumps(pipe_kwargs)

    # Ensure presence of necessary keyword arguments / widgets
    widgets = ("team", "machine", "service", "issue", "model_id", "pipe_kwargs")
    missing = set(widgets) - set(notebook_params)
    if missing:
        msg = f"Missing keys for the parameter ``notebook_params``: {missing}"
        raise ValueError(msg)

    # Send request
    api = DatabricksJobsAPI.from_os_env(host_key, access_token_key)
    return api.run_job(job_id, idempotency_token, notebook_params)



#  Copyright (c) 2022 by Amplo.

import json

import pytest
from requests import HTTPError

from amplo.api.databricks import DatabricksJobsAPI, train_on_cloud

DUMMY_JOB_ID = 749905062420360


class TestDatabricksJobsAPI:
    @classmethod
    def setup_class(cls):
        cls.api = DatabricksJobsAPI.from_os_env()

    def test_request(self):
        # A valid request should return a dictionary
        out = self.api.request("get", "2.1/jobs/list")
        assert isinstance(out, dict)
        # An invalid request should raise an error
        with pytest.raises(HTTPError):
            self.api.request("get", "")

    def test_list_jobs(self):
        expected_out = self.api.request("get", "2.1/jobs/list", {"limit": 1})
        actual_out = self.api.list_jobs(limit=1)
        assert expected_out == actual_out

    def test_list_runs(self):
        expected_out = self.api.request("get", "2.1/jobs/runs/list", {"limit": 1})
        actual_out = self.api.list_runs(limit=1)
        assert expected_out == actual_out

    def test_get_job(self):
        expected_out = self.api.request("get", "2.1/jobs/get", {"job_id": DUMMY_JOB_ID})
        actual_out = self.api.get_job(DUMMY_JOB_ID)
        assert expected_out == actual_out

    def test_get_run(self):
        # Find an existing run
        runs = self.api.list_runs(limit=1).get("runs")
        if len(runs) == 0:
            pytest.skip("There exists no run.")
        run_id = runs[0]["run_id"]

        # Test
        expected_out = self.api.request("get", "2.1/jobs/runs/get", {"run_id": run_id})
        actual_out = self.api.get_run(run_id)
        assert expected_out == actual_out

    def test_run_job(self):
        run_id = self.api.run_job(DUMMY_JOB_ID)["run_id"]
        assert self.api.cancel_run(run_id) == {}


def test_train_on_cloud():
    api = DatabricksJobsAPI.from_os_env()

    # Case 1: Missing keys
    with pytest.raises(ValueError):
        train_on_cloud(DUMMY_JOB_ID, notebook_params={})

    # Case 2: Key "pipe_kwargs" is not provided but shall be set by `train_on_cloud`
    params = {"team": 0, "machine": 0, "service": 0, "issue": 0, "model_id": 0}
    job = train_on_cloud(DUMMY_JOB_ID, notebook_params=params)
    run = api.get_run(job["run_id"])
    api.cancel_run(run["run_id"])

    passed_params = run["overriding_parameters"]["notebook_params"]
    assert set(params) | {"pipe_kwargs"} == set(passed_params)

    # Case 3: Key "pipe_kwargs" is present but default parameters shall be set
    params["pipe_kwargs"] = {"verbose": 2}
    job = train_on_cloud(DUMMY_JOB_ID, notebook_params=params)
    run = api.get_run(job["run_id"])
    api.cancel_run(run["run_id"])

    passed_params = run["overriding_parameters"]["notebook_params"]
    passed_params["pipe_kwargs"] = json.loads(passed_params["pipe_kwargs"])
    assert params["pipe_kwargs"]["verbose"] == passed_params["pipe_kwargs"]["verbose"]
    assert params.pop("pipe_kwargs") != passed_params.pop("pipe_kwargs")
    params = {key: str(params[key]) for key in params}  # requests stringifies integers
    assert params == passed_params



#  Copyright (c) 2022 by Amplo.

from amplo.classification._base import BaseClassifier
from amplo.classification.catboost import CatBoostClassifier
from amplo.classification.lgbm import LGBMClassifier
from amplo.classification.partial_boosting import PartialBoostingClassifier
from amplo.classification.stacking import StackingClassifier
from amplo.classification.xgb import XGBClassifier



#  Copyright (c) 2022 by Amplo.

import json
import os
from pathlib import Path
from typing import Union
from warnings import warn

import pandas as pd

__all__ = [
    "boolean_input",
    "parse_json",
    "read_pandas",
    "merge_logs",
    "get_log_metadata",
]


FILE_READERS = {
    ".csv": pd.read_csv,
    ".json": pd.read_json,
    ".xml": pd.read_xml,
    ".feather": pd.read_feather,
    ".parquet": pd.read_parquet,
    ".stata": pd.read_stata,
    ".pickle": pd.read_pickle,
}


def boolean_input(question: str) -> bool:
    x = input(question + " [y / n]")
    if x.lower() == "n" or x.lower() == "no":
        return False
    elif x.lower() == "y" or x.lower() == "yes":
        return True
    else:
        warn('Sorry, I did not understand. Please answer with "n" or "y"')
        return boolean_input(question)


def parse_json(json_string: Union[str, dict]) -> Union[str, dict]:
    if isinstance(json_string, dict):
        return json_string
    else:
        try:
            return json.loads(
                json_string.replace("'", '"')
                .replace("True", "true")
                .replace("False", "false")
                .replace("nan", "NaN")
                .replace("None", "null")
            )
        except json.decoder.JSONDecodeError:
            warn(f"Cannot validate, impassable JSON: {json_string}")
            return json_string


def read_pandas(path: Union[str, Path]) -> pd.DataFrame:
    """
    Wrapper for various read functions

    Returns
    -------
    pd.DataFrame
    """
    file_extension = Path(path).suffix
    if file_extension not in FILE_READERS:
        raise NotImplementedError(f"File format {file_extension} not supported.")
    else:
        reader = FILE_READERS[file_extension]
        return reader(path, low_memory=False)


def merge_logs(path_to_folder, target="labels"):
    r"""
    Combine log files from given directory into a multi-indexed dataframe

    Notes
    -----
    Make sure that each protocol is located in a sub folder whose name represents the
    respective label.

    A directory structure example:
        |   ``path_to_folder``
        |   ``├─ Label_1``
        |   ``│   ├─ Log_1.*``
        |   ``│   └─ Log_2.*``
        |   ``├─ Label_2``
        |   ``│   └─ Log_3.*``
        |   ``└─ ...``

    Parameters
    ----------
    path_to_folder : str or Path
        Parent directory
    target : str
        Column name for target

    Returns
    -------
    data : pd.DataFrame
        All logs concatenated into one multi-indexed dataframe.
        Multi-index names are ``log`` and ``index``.
        Target column depicts the folder name.
    metadata : dict
        File metadata

    Warns
    --------
    EmptyDataError
        Whenever an empty file is found.
    NotImplementedError
        Whenever a file with an unknown format is found.
    """
    # Tests
    if not Path(path_to_folder).is_dir():
        raise ValueError(f"The provided path is no directory: {path_to_folder}")
    if not Path(path_to_folder).exists():
        raise FileNotFoundError(f"Directory does not exist: {path_to_folder}")
    if not isinstance(target, str) or target == "":
        raise ValueError("Target name must be a non-empty string.")

    # Result init
    data = []

    # Get file names
    metadata = get_log_metadata(path_to_folder)

    # Loop through file paths in metadata
    for file_id in metadata:
        # Read data
        try:
            datum = read_pandas(metadata[file_id]["full_path"])
        except pd.errors.EmptyDataError:
            warn(f"Empty file: {metadata[file_id]}")
            continue
        except NotImplementedError:
            warn(f"Unknown file format: {metadata[file_id]}")

        # Set labels
        datum[target] = metadata[file_id]["folder"]

        # Set index
        datum.set_index(
            pd.MultiIndex.from_product(
                [[file_id], datum.index.values], names=["log", "index"]
            ),
            inplace=True,
        )

        # Add to list
        data.append(datum)

    if len(data) == 1:
        # Omit concatenation when only one item
        return data[0], metadata
    else:
        # Concatenate dataframes
        return pd.concat(data), metadata


def get_log_metadata(path_to_folder):
    """Get metadata of log files

    Parameters
    ----------
    path_to_folder

    Notes
    -----
    Make sure that each protocol is located in a sub folder whose name represents the
    respective label.

    A directory structure example:
        |   ``path_to_folder``
        |   ``├─ Label_1``
        |   ``│   ├─ Log_1.*``
        |   ``│   └─ Log_2.*``
        |   ``├─ Label_2``
        |   ``│   └─ Log_3.*``
        |   ``└─ ...``

    Returns
    -------
    metadata : dict
        Dictionary whose keys depict the file id (integer) and each value contains a
        dictionary with ``folder``, ``file``, ``full_path`` and ``last_modified`` key.
    """
    # Checks
    if not Path(path_to_folder).is_dir():
        raise ValueError(f"The provided path is no directory: {path_to_folder}")
    if not Path(path_to_folder).exists():
        raise FileNotFoundError(f"Directory does not exist: {path_to_folder}")

    # Init
    metadata = dict()
    file_id = 0

    # Loop through folders
    for folder in sorted(Path(path_to_folder).iterdir()):

        # Loop through files (ignore hidden files)
        for file in sorted(folder.glob("[!.]*.*")):

            # Check file
            if file.suffix not in FILE_READERS:
                warn(f"Skipped unsupported file format: {file}")
                continue
            elif file.stat().st_size == 0:
                warn(f"Skipped empty file: {file}")
                continue

            # Add to metadata
            metadata[file_id] = {
                "folder": str(folder.name),
                "file": str(file.name),
                "full_path": str(file.resolve()),
                "last_modified": os.path.getmtime(str(file)),
            }

            # Increment
            file_id += 1

    if file_id == 0:
        raise FileNotFoundError(
            "Directory seems to be empty. Check whether you specified the correct path."
        )

    return metadata



#  Copyright (c) 2022 by Amplo.

import pandas as pd
import pytest
from sklearn.datasets import make_classification, make_regression

from amplo.automl import Modeller
from tests import rmtree


@pytest.fixture(scope="class", params=["classification", "regression"])
def make_mode(request):
    mode = request.param
    if mode == "classification":
        x, y = make_classification()
        objective = "neg_log_loss"
    elif mode == "regression":
        x, y = make_regression()
        objective = "r2"
    else:
        raise ValueError("Invalid mode")
    request.cls.mode = mode
    request.cls.objective = objective
    request.cls.x = pd.DataFrame(x)
    request.cls.y = pd.Series(y)
    yield


@pytest.fixture(scope="class", autouse=True)
def clean():
    yield None
    rmtree("tmp/")


@pytest.mark.usefixtures("make_mode")
class TestModelling:
    folder = "tmp/"

    def test_modeller(self):
        mod = Modeller(mode=self.mode, objective=self.objective, folder=self.folder)
        mod.fit(self.x, self.y)
        # Tests
        assert isinstance(
            mod.results, pd.DataFrame
        ), "Results should be type pd.DataFrame"
        assert len(mod.results) != 0, "Results empty"
        assert mod.results["mean_objective"].max() < 1, "R2 needs to be smaller than 1"
        assert (
            not mod.results["mean_objective"].isna().any()
        ), "Mean Objective shouldn't contain NaN"
        assert (
            not mod.results["std_objective"].isna().any()
        ), "Std Objective shouldn't contain NaN"
        assert (
            not mod.results["mean_time"].isna().any()
        ), "Mean time shouldn't contain NaN"
        assert (
            not mod.results["std_time"].isna().any()
        ), "Std time shouldn't contain NaN"
        assert "date" in mod.results.keys()
        assert "model" in mod.results.keys()
        assert "dataset" in mod.results.keys()
        assert "params" in mod.results.keys()

    @pytest.mark.parametrize("n_samples", [100, 100_000])
    def test_return(self, n_samples):
        Modeller(mode=self.mode, samples=n_samples).return_models()



#  Copyright (c) 2022 by Amplo.

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification

from amplo.automl import DataSampler


class TestDataSampler:
    def test_binary_small_under(self):
        print("Binary Small Under")
        # Data
        x, y = make_classification(n_classes=2, n_samples=100, weights=[0.05, 0.95])
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="under")
        x_res, y_res = sampler.fit_resample(x, y)

        # Tests
        assert len(y_res) <= len(y), "y_res longer than y"

    def test_binary_small_over(self):
        print("Binary Small Over")
        # Data
        x, y = make_classification(n_classes=2, n_samples=1000, weights=[0.05, 0.95])
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="over")
        x_res, y_res = sampler.fit_resample(x, y)

        # Tests
        assert len(y_res) >= len(y), "y_res shorter than y"

    def test_binary_big_under(self):
        print("Binary Big Under")
        # Data
        x, y = make_classification(n_classes=2, n_samples=100000, weights=[0.05, 0.95])
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="under")
        x_res, y_res = sampler.fit_resample(x, y)

        # Tests
        assert len(y_res) <= len(y), "y_res longer than y"

    def test_binary_big_over(self):
        print("Binary Big Over")
        # Data
        x, y = make_classification(n_classes=2, n_samples=100000, weights=[0.05, 0.95])
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="over")
        x_res, y_res = sampler.fit_resample(x, y)

        # Tests
        assert len(y_res) >= len(y), "y_res shorter than y"
        assert sampler.opt_sampler is None, "Shouldn't over-sample big data"

    def test_multi_small_under(self):
        print("Multi Small Under")
        # Data
        x, y = make_classification(
            n_classes=5,
            n_features=10,
            n_informative=10,
            n_redundant=0,
            n_repeated=0,
            n_clusters_per_class=1,
            n_samples=100,
            weights=[0.05, 0.2, 0.1, 0.05, 0.6],
        )
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="under")
        x_res, y_res = sampler.fit_resample(x, y)

        # Tests
        assert len(y_res) <= len(y), "y_res longer than y"
        assert sampler.opt_sampler is None, "Shouldn't under-sample on small data"

    def test_multi_small_over(self):
        print("Multi Small Over")
        # Data
        x, y = make_classification(
            n_classes=5,
            n_features=10,
            n_informative=10,
            n_redundant=0,
            n_repeated=0,
            n_clusters_per_class=1,
            n_samples=100,
            weights=[0.05, 0.2, 0.1, 0.05, 0.6],
        )
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="over")
        x_res, y_res = sampler.fit_resample(x, y)

        # Tests
        assert len(y_res) >= len(y), "y_res shorter than y"

    def test_multi_big_under(self):
        print("Multi Big Under")
        # Data
        x, y = make_classification(
            n_classes=5,
            n_features=10,
            n_informative=10,
            n_redundant=0,
            n_repeated=0,
            n_clusters_per_class=1,
            n_samples=100000,
            weights=[0.05, 0.2, 0.1, 0.05, 0.6],
        )
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="under")
        x_res, y_res = sampler.fit_resample(x, y)

        # Tests
        assert len(y_res) <= len(y), "y_res longer than y"

    def test_multi_big_over(self):
        print("Multi Big Over")
        # Data
        x, y = make_classification(
            n_classes=5,
            n_features=10,
            n_informative=10,
            n_redundant=0,
            n_repeated=0,
            n_clusters_per_class=1,
            n_samples=100000,
            weights=[0.05, 0.2, 0.1, 0.05, 0.6],
        )
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="over")
        x_res, y_res = sampler.fit_resample(x, y)

        # Tests
        assert len(y_res) <= len(y), "y_res longer than y"
        assert sampler.opt_sampler is None, "Shouldn't over-sample on big data"

    def test_balanced(self):
        print("Balanced")
        # Data
        x, y = make_classification(n_samples=100)
        x, y = pd.DataFrame(x), pd.Series(y)

        # Resample
        sampler = DataSampler(method="both")
        x_res, y_res = sampler.fit_resample(x, y)

        assert np.allclose(x_res, x), "X is not the same"
        assert sampler.opt_sampler is None, "Classes are balanced"



#  Copyright (c) 2022 by Amplo.

import copy
import math
import os
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from sklearn import metrics
from sklearn.model_selection import KFold

from amplo.documenting.binary import BinaryDocumenting
from amplo.utils.logging import logger
from amplo.utils.util import deprecated

if TYPE_CHECKING:
    from amplo import Pipeline


@deprecated(
    "This is no longer used, all information is available in the created settings.json."
)
class RegressionDocumenting(BinaryDocumenting):
    def __init__(self, pipeline: "Pipeline"):
        super().__init__(pipeline)
        self.y_not_standardized = None

    def analyse(self):
        # Cross-Validation Plots
        fig, ax = plt.subplots(
            math.ceil(self.p.cv_splits / 2),
            2,
            sharex="all",
            sharey="all",
            figsize=[24, 8],
        )
        fig.suptitle(
            "{}-Fold Cross Validated Predictions - {}".format(
                self.p.cv_splits, self.mName
            )
        )

        # Initialize iterables
        mae = []
        mse = []
        r2 = []
        max_error = []
        rel_error = []
        self.cv = KFold(n_splits=self.p.cv_splits, shuffle=self.p.shuffle)
        # Cross Validate
        i = 0
        for i, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            xt, xv, yt, yv = (
                self.x[t],
                self.x[v],
                self.y[t].reshape((-1)),
                self.y[v].reshape((-1)),
            )
            model = copy.deepcopy(self.model)
            model.fit(xt, yt)
            prediction = model.predict(xv)

            # Metrics
            mae.append(metrics.mean_absolute_error(yv, prediction))
            mse.append(metrics.mean_squared_error(yv, prediction))
            r2.append(metrics.r2_score(yv, prediction))
            max_error.append(metrics.max_error(yv, prediction))
            rel_error.append(metrics.mean_absolute_percentage_error(yv, prediction))

            # Plot
            ax[i // 2][i % 2].set_title("Fold-{}".format(i))
            if self.p.standardize:
                ax[i // 2][i % 2].plot(self.y_not_standardized[v], color="#2369ec")
                ax[i // 2][i % 2].plot(
                    self.p.bestOutputScaler.inverse_transform(model.predict(xv)),
                    color="#ffa62b",
                    alpha=0.4,
                )
            else:
                ax[i // 2][i % 2].plot(yv, color="#2369ec")
                ax[i // 2][i % 2].plot(model.predict(xv), color="#ffa62b", alpha=0.4)

        # Store
        self.metrics = {
            "Mean Absolute Error": [float(np.mean(mae)), float(np.std(mae))],
            "Mean Squared Error": [float(np.mean(mse)), float(np.std(mse))],
            "R2 Score": [float(np.mean(r2)), float(np.std(r2))],
            "Max Error": [float(np.mean(max_error)), float(np.std(max_error))],
            "Mean Relative Error": [
                float(np.mean(rel_error)),
                float(np.std(rel_error)),
            ],
        }
        self.outputMetrics = copy.deepcopy(self.metrics)

        # Save figure
        ax[i // 2][i % 2].legend(["Output", "Prediction"])
        if not os.path.exists(
            self.p.main_dir + "EDA/Validation/v{}/".format(self.p.version)
        ):
            os.makedirs(self.p.main_dir + "EDA/Validation/v{}/".format(self.p.version))
        cross_val_path = self.p.main_dir + "EDA/Validation/v{}/Cross_Val_{}.png".format(
            self.p.version, self.mName
        )
        fig.savefig(cross_val_path, format="png", dpi=200)

        # Print & Finish plot
        logger.info(
            "Mean Absolute Error:            {:.2f} \u00B1 {:.2f}".format(
                np.mean(mae), np.std(mae)
            )
        )
        logger.info(
            "Mean Squared Error:             {:.2f} \u00B1 {:.2f}".format(
                np.mean(mse), np.std(mse)
            )
        )
        logger.info(
            "R2 Score:                       {:.2f} \u00B1 {:.2f}".format(
                np.mean(r2), np.std(r2)
            )
        )
        logger.info(
            "Max Error:                      {:.2f} \u00B1 {:.2f}".format(
                np.mean(max_error), np.std(max_error)
            )
        )
        logger.info(
            "Mean Absolute Percentage Error: {:.2f} \u00B1 {:.2f}".format(
                np.mean(rel_error), np.std(rel_error)
            )
        )

        # Feature Importance
        if not os.path.exists(
            self.p.main_dir + "EDA/Features/v{}/RF.png".format(self.p.version)
        ):
            if not os.path.exists(
                self.p.main_dir + "EDA/Features/v{}".format(self.p.version)
            ):
                os.makedirs(
                    self.p.main_dir + "EDA/Features/v{}/".format(self.p.version)
                )
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines["right"].set_visible(False)
            ax.spines["bottom"].set_visible(False)
            ax.spines["top"].set_visible(False)
            fi = self.p.feature_processor.feature_importance_["rf"]
            plt.barh(list(fi)[:15], width=list(fi.values())[:15], color="#2369ec")
            fig.savefig(
                self.p.main_dir + "EDA/Features/v{}/RF.png".format(self.p.version),
                format="png",
                dpi=200,
            )

    def model_performance(self):
        if not self.check_new_page():
            self.ln(self.lh)
        self.add_h2("Model Performance")
        self.add_text(
            "Model performance is analysed by various metrics. This model has been "
            "selected based on the {} score.".format(self.p.objective)
        )

        # Metrics
        self.set_font("Helvetica", "B", 12)
        self.ln(self.lh)
        self.cell(w=50, h=self.lh, txt="Metric", border="B", align="C")
        self.cell(w=50, h=self.lh, txt="Score", border="LB", align="C")
        self.set_font("Helvetica", "", 12)
        for k, v in self.metrics.items():
            self.ln(self.lh)
            self.cell(w=50, h=self.lh, txt=k, border="R", align="L")
            self.cell(
                w=50,
                h=self.lh,
                txt=f"{v[0]:.2f} \u00B1 {v[1]:.2f}",
                border="L",
                align="C",
            )
        self.ln(self.lh * 3)

    def validation(self):
        if not self.check_new_page():
            self.ln(self.lh)

        self.add_h3("Cross Validation Plot")
        x, y = self.get_x(), self.get_y()
        path = self.p.main_dir + "EDA/Validation/v{}/Cross_Val_{}.png".format(
            self.p.version, self.mName
        )
        self.image(x=(self.WIDTH - 200) / 2, y=y, w=200, h=90, name=path)
        self.add_page()

        self.add_h3("Validation Strategy")
        self.add_text(
            "All experiments are cross-validated. This means that every time a model's "
            "performance is evaluated, it's trained on one part of the data, and test "
            "on another. Therefore, the model is always test against data it has not "
            "yet been trained for. This gives the best approximation for real world "
            "(out of sample) performance. The current validation strategy used "
            "is {}, with {} splits and {} shuffling the data.".format(
                type(self.cv).__name__,
                self.p.cv_splits,
                "with" if self.p.shuffle else "without",
            )
        )



#  Copyright (c) 2022 by Amplo.

"""
Implements base classes.
"""

import importlib
import inspect
import logging
from collections import defaultdict
from warnings import warn

from amplo.base.exceptions import NotFittedError
from amplo.utils.logging import logger

__all__ = [
    "BaseObject",
    "BaseEstimator",
    "BaseTransformer",
    "BasePredictor",
    "LoggingMixin",
]


class BaseObject:
    # This class is strongly inspired by sklearn's BaseEstimator.
    """
    Object base class.

    Class attributes
    ----------------
    _add_to_settings : list of str
        Attribute names to be included in settings.
    """

    _add_to_settings = []

    def get_settings(self, deep=True):
        """
        Get setting parameters for this object.

        Parameters
        ----------
        deep : bool
            If True, will return the parameters for this object and contained
            sub-objects.

        Returns
        -------
        settings : dict
            Settings for this object.
        """
        out = {"<init_params>": self.get_params(deep=False)}
        for key in self._add_to_settings:
            value = getattr(self, key, None)
            if deep and isinstance(value, BaseObject):
                deep_items = value.get_settings().items()
                out.update((f"{key}__{k}", val) for k, val in deep_items)
                out[key] = str(type(value))
            else:
                out[key] = value
        return out

    def load_settings(self, settings):
        """
        Load settings for this object.

        Parameters
        ----------
        settings : dict

        Returns
        -------
        self :
            An object with injected settings.
        """
        if not settings:
            return self
        settings = settings.copy()
        valid_settings = self.get_settings(deep=True)

        if "<init_params>" in settings:
            self.__init__(**settings.pop("<init_params>"))  # noqa

        nested_settings = defaultdict(dict)  # grouped by prefix
        for key, value in settings.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_settings:
                local_valid_params = self._add_to_settings
                raise ValueError(
                    f"Invalid parameter {key!r} for object {self}. "
                    f"Valid parameters are: {local_valid_params!r}."
                )

            if delim:
                nested_settings[key][sub_key] = value
            else:
                if isinstance(value, str) and "<class " in value:
                    _, cls, _ = value.split("'")
                    module_name, class_name = cls.rsplit(".", 1)
                    module = importlib.import_module(module_name)
                    init_params = settings.get(f"{key}__<init_params>", {})
                    value = getattr(module, class_name)(**init_params)
                setattr(self, key, value)
                valid_settings[key] = value

        for key, sub_params in nested_settings.items():
            valid_settings[key].load_settings(sub_params)

        return self

    @classmethod
    def _get_param_names(cls):
        """
        Get parameter names for the estimator.
        """
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [
            p
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError(
                    "scikit-learn estimators should always "
                    "specify their parameters in the signature"
                    " of their __init__ (no varargs)."
                    " %s with constructor %s doesn't "
                    " follow this convention." % (cls, init_signature)
                )
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key, None)
            if deep and hasattr(value, "get_params"):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        """
        Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as :class:`~sklearn.pipeline.Pipeline`). The latter have
        parameters of the form ``<component>__<parameter>`` so that it's
        possible to update each component of a nested object.

        Parameters
        ----------
        **params : dict
            Estimator parameters.

        Returns
        -------
        self
            Estimator instance.
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                local_valid_params = self._get_param_names()
                raise ValueError(
                    f"Invalid parameter {key!r} for estimator {self}. "
                    f"Valid parameters are: {local_valid_params!r}."
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

    def reset(self):
        """
        Reset the object to a clean post-init state.

        Equivalent to sklearn.clone but overwrites self.
        After self.reset() call, self is equal in value to
        `type(self)(**self.get_params(deep=False))`

        Detail behaviour:
            1. removes any object attributes, except:
                - hyperparameters = arguments of __init__
                - object attributes containing double-underscores, i.e. "__"
            2. runs __init__ with current values of hyperparameters (result of
            get_params)

        Not affected by the reset are:
        - object attributes containing double-underscores
        - class and object methods, class attributes
        """
        # retrieve parameters to copy them later
        params = self.get_params(deep=False)

        # delete all object attributes in self
        attrs = [attr for attr in dir(self) if "__" not in attr]
        cls_attrs = [attr for attr in dir(type(self))]
        self_attrs = set(attrs).difference(cls_attrs)
        for attr in self_attrs:
            delattr(self, attr)

        # run init with a copy of parameters self had at the start
        self.__init__(**params)  # noqa

        return self


class BaseEstimator(BaseObject):
    """
    Estimator base class.

    Extends the BaseObject class with an is_fitted attribute.

    Attributes
    ----------
    _is_fitted : bool
        Indicates whether the estimator is fitted.
    """

    _add_to_settings = ["_is_fitted", *BaseObject._add_to_settings]

    def __init__(self):
        self._is_fitted = False
        super().__init__()

    @property
    def is_fitted(self):
        """Whether `fit` has been called."""
        # Note: for backwards compatibility
        if hasattr(self, "_is_fitted"):
            return self._is_fitted
        else:
            return True

    def check_is_fitted(self):
        """
        Asserts that estimator is fitted.

        Raises
        ------
        NotFittedError
            When estimator is not fitted.
        """
        if not self.is_fitted:
            raise NotFittedError(
                f"This instance of {self.__class__.__name__} has not been "
                f"fitted yet; please call `fit` first."
            )

    def fit(self, x, y=None, **fit_params):
        """
        Fit the estimator to x and optionally to y.

        State change:
            Changes state to "fitted".

        Writes to self:
            Sets is_fitted flag to True.
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Feature data to fit.
        y : numpy.ndarray or pandas.Series
            Target data to fit.
        **fit_params : dict
            Additional fit parameters.

        Returns
        -------
        self : estimator
            A fitted instance of the estimator.
        """
        # If fit is called, estimator is reset, including fitted state
        self.reset()

        # Pass to inner fit
        self._fit(x, y, **fit_params)

        # This should happen last: fitted state is set to True
        self._is_fitted = True

        return self

    def _fit(self, x, y=None, **fit_params):
        """
        Fit the estimator to x and optionally to y.

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Checked feature data to fit.
        y : numpy.ndarray or pandas.Series
            Checked target data to fit.
        **fit_params : dict
            Additional fit parameters.

        Returns
        -------
        self : estimator
            A fitted instance of the transformer.
        """
        return self


class BaseTransformer(BaseEstimator):
    """
    Transformer base class.
    """

    def transform(self, x, y=None):
        """
        Transform data and return it.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_".
            self._is_fitted

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Feature data to transform.
        y : numpy.ndarray or pandas.Series
            Additional target data to transform.

        Returns
        -------
        pandas.DataFrame
            Transformed version of x.
        """
        # Check whether is fitted
        self.check_is_fitted()

        # Transform data
        xt = self._transform(x, y)

        return xt

    def fit_transform(self, x, y=None, **fit_params):
        """
        Fit and transform data.

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Feature data to transform.
        y : numpy.ndarray or pandas.Series
            Additional target data to transform.
        **fit_params : dict
            Additional fit parameters.

        Returns
        -------
        pandas.DataFrame
            Transformed version of x.
        """
        # Default, non-optimized version for `fit_transform`. Overwrite, when
        # can be optimized.
        return self.fit(x, y, **fit_params).transform(x, y)

    def _transform(self, x, y=None):
        """
        Transform data.

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Checked feature data to fit.
        y : numpy.ndarray or pandas.Series
            Checked target data to fit.

        Returns
        -------
        pandas.DataFrame
            Transformed version of x.
        """
        raise NotImplementedError("Abstract method.")


class BasePredictor(BaseEstimator):
    """
    Predictor base class.
    """

    def predict(self, x, y=None, **predict_params):
        """
        Transform data and return it.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_".
            self._is_fitted

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Feature data to transform.
        y : numpy.ndarray or pandas.Series
            Additional target data to transform.
        **predict_params
            Additional predict parameters.

        Returns
        -------
        pandas.DataFrame
            Prediction for x.
        """
        # Check whether is fitted
        self.check_is_fitted()

        # Transform data
        xt = self._predict(x, y, **predict_params)

        return xt

    def fit_predict(self, x, y=None, **fit_predict_params):
        """
        Fit and predict data.

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Feature data to transform.
        y : numpy.ndarray or pandas.Series
            Additional target data to transform.
        **fit_predict_params : dict
            Additional fit and predict parameters.

        Returns
        -------
        pandas.DataFrame
            Prediction for x.
        """
        # Default, non-optimized version for `fit_predict`. Overwrite, when
        # can be optimized.
        return self.fit(x, y, **fit_predict_params).predict(x, **fit_predict_params)

    def _predict(self, x, y=None, **predict_params):
        """
        Predict on data.

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Checked feature data to fit.
        y : numpy.ndarray or pandas.Series
            Checked target data to fit.
        **predict_params : dict
            Additional predict parameters.

        Returns
        -------
        pandas.DataFrame
            Prediction for x.
        """
        raise NotImplementedError("Abstract method.")


class LoggingMixin:
    """
    Mixin class for adding logging capability to an object.

    Parameters
    ----------
    verbose : int
        Verbosity for logger.

    Notes
    -----
    The logging level depends on the parameter verbose as follows:
        - verbose=0: warnings or higher priority
        - verbose=1: info or higher priority
        - verbose=2: debugging info or higher priority
    """

    def __init__(self, verbose=0):
        if not isinstance(verbose, (float, int)):
            raise ValueError(f"Invalid dtype for `verbose`: {type(verbose)}.")

        # Set logging level based on verbose
        if verbose < 0:
            warn("`verbose` cannot be smaller than zero.", UserWarning)
            verbose = 0
            logging_level = logging.WARNING
        elif verbose == 0:
            logging_level = logging.WARNING
        elif verbose == 1:
            logging_level = logging.INFO
        else:  # verbose >= 2
            logging_level = logging.DEBUG

        self.verbose = verbose
        # Notice: Without creating a new logger (through `getChild`), setting the
        # logging level will influence all logging. Setting logging levels individually
        # per class is therefore not possible.
        self.logger = logger.getChild(type(self).__name__)
        self.logger.setLevel(logging_level)
